import configparser
import pickle
from pathlib import Path

import click
import requests

from kks.ejudge import AuthData, ejudge_auth, check_session


def config_directory():
    directory = Path(click.get_app_dir('kks', force_posix=True))
    directory.mkdir(exist_ok=True)
    return directory


def read_config():
    config = configparser.ConfigParser()
    config.optionxform = str
    cfg = config_directory() / 'config.ini'
    if cfg.is_file():
        config.read(cfg)
    return config


def write_config(config):
    cfg = config_directory() / 'config.ini'
    with cfg.open('w') as f:
        config.write(f)


def store_session(session):
    cookies = config_directory() / 'cookies.pickle'
    with open(cookies, 'wb') as f:
        pickle.dump(session.cookies, f)


def load_session():
    cookies = config_directory() / 'cookies.pickle'
    if not cookies.is_file():
        return None

    session = requests.session()
    with open(cookies, 'rb') as f:
        try:
            cookies = pickle.load(f)
        except Exception:
            return None
        session.cookies.update(cookies)
    return session


def save_auth_data(auth_data, store_password=True):
    config = read_config()
    config['Auth'] = {
        'login': auth_data.login,
        'contest': auth_data.contest_id
    }

    if store_password and auth_data.password is not None:
        config['Auth']['password'] = auth_data.password

    write_config(config)


def load_auth_data():
    config = read_config()
    if not config.has_section('Auth'):
        return None
    auth = config['Auth']
    if 'login' in auth and 'contest' in auth:
        return AuthData(auth['login'], auth['contest'], auth.get('password', None))
    return None


def save_links(links):
    config = read_config()
    config['Links'] = links
    write_config(config)


def load_links():
    config = read_config()
    if config.has_section('Links'):
        return config['Links']
    return None


def get_valid_session():
    session = load_session()

    if session is not None:
        links = load_links()
        if links is None or not check_session(links, session):
            session = None

    if session is None:
        auth_data = load_auth_data()
        if auth_data is None:
            click.secho('No valid cookies or auth data, please use "kks auth" to log in', fg='yellow', err=True)
            return None

        click.secho('Cookies are either missing or invalid, trying to auth with saved data', fg='yellow', err=True)
        store_password = True
        if auth_data.password is None:
            store_password = False
            auth_data.password = click.prompt('Password', hide_input=True)

        session = requests.session()
        links = ejudge_auth(auth_data, session)
        if links is None:
            return None

        save_auth_data(auth_data, store_password)
        save_links(links)
        store_session(session)

    return session


def get_clang_style_string():
    cfg = config_directory() / '.clang-format'
    if cfg.exists():
        with cfg.open('r') as f:
            return f.read()

    return r"""{
        Language: Cpp,
        BasedOnStyle: Google,
        IndentWidth: 4,
        UseTab: Never,
        NamespaceIndentation: All,
        ColumnLimit: 80,
        AccessModifierOffset: -4,
        AlignAfterOpenBracket: AlwaysBreak,
        AlignOperands: false,
        AlwaysBreakTemplateDeclarations: Yes,
        BinPackArguments: false,
        BinPackParameters: false,
        AllowShortFunctionsOnASingleLine: Empty,
        BreakBeforeBraces: Custom,
        BraceWrapping: { AfterEnum: true, AfterStruct: true }
    }"""


def find_workspace(path=None):
    if path is None:
        path = Path()

    path = path.resolve()

    while path.is_dir():
        file = path / '.kks-workspace'
        if file.exists():
            return path
        if path == path.parent:
            return None
        path = path.parent


def get_solution_directory():
    workspace = find_workspace()

#     if contest is not None and task is not None:
#         if workspace is not None:
#             result = workspace
#
#             if not result.is_dir():
#                 click.secho(f'Path {result} is not a directory', fg='red', err=True)
#                 return None
#
#             return result
#         else:
#             click.secho('Could not find workspace', fg='red', err=True)
#             return None

    return Path().absolute()



import click
import requests
from bs4 import BeautifulSoup


class LinkTypes:
    SETTINGS = 'Settings'
    SUMMARY = 'Summary'
    SUBMISSIONS = 'Submissions'
    USER_STANDINGS = 'User standings'
    SUBMIT_CLAR = 'Submit clar'
    CLARS = 'Clars'


class Status:
    OK = 'OK'
    REVIEW = 'Pending review'
    NOT_SUBMITTED = 'Not submitted'


class AuthData:
    def __init__(self, login, contest_id, password=None):
        self.login = login
        self.contest_id = contest_id
        self.password = password


class Problem:
    def __init__(self, short_name, name, href, status, tests_passed, score):
        self.short_name = short_name
        self.name = name
        self.href = href
        self.status = status
        self.tests_passed = tests_passed
        self.score = score


def ejudge_auth(auth_data, session):
    url = f'https://caos.ejudge.ru/ej/client?contest_id={auth_data.contest_id}'

    page = session.post(url, data={
        'login': auth_data.login,
        'password': auth_data.password
    })

    if page.status_code != requests.codes.ok:
        click.secho(f'Failed to authenticate (status code {page.status_code})', err=True, fg='red')

    soup = BeautifulSoup(page.content, 'html.parser')

    if 'Invalid contest' in soup.text or 'invalid contest_id' in soup.text:
        click.secho(f'Invalid contest (contest id {auth_data.contest_id})', fg='red', err=True)
        return None

    if 'Permission denied' in soup.text:
        click.secho('Permission denied (invalid username, password or contest id)', fg='red', err=True)
        return None

    buttons = soup.find_all('a', {'class': 'menu'}, href=True)

    return {
        button.text: button['href']
        for button in buttons
    }


def ejudge_summary(links, session):
    summary = links.get(LinkTypes.SUMMARY, None)
    if summary is None:
        return None

    page = session.get(summary)
    soup = BeautifulSoup(page.content, 'html.parser')

    tasks = soup.find_all('td', {'class': 'b1'})

    problems = []
    for problem in chunks(tasks, 6):
        short, name, status, tests_passed, score, _ = problem

        problems.append(Problem(
            short.text,
            name.text,
            name.a['href'],
            status.text if not status.text.isspace() else Status.NOT_SUBMITTED,
            tests_passed.text if not tests_passed.text.isspace() else None,
            score.text if not score.text.isspace() else 0
        ))

    return problems


def ejudge_sample(problem_link, session):
    page = session.get(problem_link)
    soup = BeautifulSoup(page.content, 'html.parser')

    input_data, output_data = None, None

    input_title = soup.find('h4', text='Input')
    if input_title is not None:
        input_data = input_title.find_next('pre').text

    output_title = soup.find('h4', text='Output')
    if output_title is not None:
        output_data = output_title.find_next('pre').text

    return input_data, output_data


def check_session(links, session):
    summary = links.get(LinkTypes.SUMMARY, None)
    if summary is None:
        return False
    response = session.get(summary)
    if 'Invalid session' in response.text:
        return False
    return True


def chunks(iterable, chunk_size):
    for i in range(0, len(iterable), chunk_size):
        yield iterable[i:i + chunk_size]



import click
import requests

from kks.ejudge import ejudge_auth, AuthData
from kks.util import store_session, save_auth_data, save_links


@click.command(short_help='Authorize and save authentication data to configuration directory')
@click.option('-l', '--login', prompt=True)
@click.password_option('-p', '--password', confirmation_prompt=False)
@click.option('-c', '--contest-id', prompt=True, type=int,
              help='''Ejudge contest id.
              For example, https://caos.ejudge.ru/ej/client?contest_id=133 has contest id 133''')
@click.option('--store-password/--no-store-password', default=True,
              help='''Toggle storing plaintext password in config for auto-login.
              If disabled, only session cookies will be stored.
              Enabled by default.''')
def auth(login, password, contest_id, store_password):
    """Authorize and save authentication data to configuration directory

    \b
    Stored files:
    - config.ini     - contains login, password and group number
    - cookies.pickle - last active ejudge session cookies
    - data.json      - data parsed from ejudge (page urls, for example)"""

    session = requests.session()
    auth_data = AuthData(login, contest_id, password)

    links = ejudge_auth(auth_data, session)
    if links is None:
        return

    click.secho('Successfully logged in', fg='green')

    save_auth_data(auth_data, store_password)
    save_links(links)
    store_session(session)

    click.secho('Successfully saved auth data', fg='green', err=True)



import subprocess

import click
from tqdm import tqdm

from kks.util import get_solution_directory


@click.command(short_help='Run solution')
@click.option('-o', '--output-only', is_flag=True,
              help='If specified, only solution will be run. Useful to generate output for manually created tests')
@click.option('-g', '--generator', type=click.Path(exists=True),
              help='Script, used to generate .in files')
@click.option('-s', '--solution', type=click.Path(exists=True),
              help='Script, used to generate .out files')
@click.option('-t', '--test', type=int,
              help='Test number to generate')
@click.option('-r', '--range', 'test_range', type=int, nargs=2,
              help='Tests to generate')
@click.option('-f', '--force', is_flag=True,
              help='Overwrite .in files')
@click.argument('gen_args', nargs=-1, type=click.UNPROCESSED)
def gen(output_only, generator, solution, test, test_range, force, gen_args):
    """
    Generate tests

    Generate script is run with arguments "TEST_NUMBER GEN_ARGS". It's output is saved into .in file.

    Solution script is run with test number as argument, and .in file as stdin. It's output is saved into .out file.

    \b
    Example usage:
      kks gen -t 1337
      kks gen -r 1 100
      kks gen -t 1 -o
      kks gen -g gen.py -s solve.py -r 1 50 -f
    """

    directory = get_solution_directory()

    generator = generator or directory / 'gen.py'
    solution = solution or directory / 'solve.py'

    if not generator.exists() and not output_only:
        click.secho(f'Generator {generator} does not exist', fg='red', err=True)
        return

    if not solution.exists():
        click.secho(f'Solution {solution} does not exist', fg='red', err=True)
        return

    if test is not None and len(test_range) != 0:
        click.secho(f'Either test or range should be specified', fg='red', err=True)
        return

    if test is not None:
        test_numbers = [test]
    else:
        l, r = sorted(test_range or (1, 100))
        test_numbers = list(range(l, r + 1))

    tests_dir = directory / 'tests'
    tests_dir.mkdir(exist_ok=True)

    t = tqdm(test_numbers, leave=False)
    for i in t:
        name = str(i).rjust(3, '0')

        input_file = tests_dir / (name + '.in')
        output_file = tests_dir / (name + '.out')

        if output_only:
            if not input_file.exists():
                click.secho('Input file ' + click.style(input_file.name, fg='blue', bold=True) +
                            ' does not exists, skipping')
                continue
        else:
            if input_file.exists() and not force:
                click.secho('Input file ' + click.style(input_file.name, fg='blue', bold=True) +
                            ' already exists, skipping. Specify -f to overwrite')
                continue

            if output_file.exists() and not force:
                click.secho('Output file ' + click.secho(output_file.name, fg='blue', bold=True) +
                            ' already exists, skipping. Specify -f to overwrite')
                continue

        t.set_description('Generating test ' + click.style(name, fg='blue', bold=True))

        if not output_only:
            with input_file.open('w') as f:
                args = [str(i)] + list(gen_args)
                process = subprocess.run(['python3', generator] + args, stdout=f)
                if process.returncode != 0:
                    click.secho('Generator exited with code ' +
                                click.style(str(process.returncode), fg='red', bold=True) +
                                ' (args: ' + ' '.join(args) + ')', fg='yellow')

        with input_file.open('r') as f_in, output_file.open('w') as f_out:
            args = [str(i)]
            process = subprocess.run(['python3', solution] + args, stdin=f_in, stdout=f_out)
            if process.returncode != 0:
                click.secho('Solution exited with code ' +
                            click.style(str(process.returncode), fg='red', bold=True) +
                            ' (args: ' + ' '.join(args) + ')', fg='yellow')

    click.secho('Generated tests!', fg='green', err=True)



import difflib
import subprocess

import click
from tqdm import tqdm

from kks.binary import compile_solution
from kks.util import get_solution_directory


@click.command(short_help='Test solutions')
@click.option('-m', '--mode', default='auto', type=click.Choice(['auto'], case_sensitive=False))
def test(mode):
    """Test solution"""

    directory = get_solution_directory()
    if directory is None:
        return

    binary = compile_solution(directory, mode)
    if binary is None:
        return

    tests_dir = directory / 'tests'

    if not tests_dir.exists():
        click.secho('No tests directory', fg='red', err=True)

    input_files = tests_dir.glob('*.in')
    input_files = sorted(input_files, key=lambda file: file.name)

    t = tqdm(input_files, leave=False)
    for input_file in t:
        test_name = input_file.stem
        output_file = tests_dir / (test_name + '.out')

        styled_file = click.style(input_file.name, fg='blue', bold=True)

        if not output_file.is_file():
            click.secho(f'No output file for test {styled_file}, skipping', fg='yellow', err=True)
            continue

        t.set_description(f'Running {styled_file}')

        if not run_test(binary, input_file, output_file):
            t.close()
            return

    click.secho('All tests passed!', fg='green')


def run_test(binary, input_file, output_file):
    with input_file.open('r') as input_f:
        process = subprocess.run(binary, stdin=input_f, capture_output=True)

    if process.returncode != 0:
        click.secho('RE', fg='red', bold=True)
        click.secho(f'Process exited with core {process.returncode}', fg='red')
        return False

    with output_file.open('r') as output_f:
        expected_output = output_f.read()

    actual_output = process.stdout.decode('utf-8')

    if expected_output != actual_output:
        click.secho('WA', fg='red', bold=True)

        expected_lines = expected_output.splitlines(keepends=True)
        actual_lines = actual_output.splitlines(keepends=True)
        diff = difflib.unified_diff(expected_lines, actual_lines, 'expected', 'actual')
        for line in diff:
            add = line.startswith('+')
            remove = line.startswith('-')
            color = 'red' if remove else 'green' if add else None
            click.secho(line, nl=False, fg=color)
        click.secho()
        return False

    return True









import subprocess
import sys

import click

from kks.binary import compile_solution
from kks.util import get_solution_directory


@click.command(short_help='Run solution')
@click.option('-m', '--mode', default='auto', type=click.Choice(['auto'], case_sensitive=False))
@click.argument('run_args', nargs=-1, type=click.UNPROCESSED)
def run(mode, run_args):
    """Run solution

    \b
    Example usage:
      kks run
      kks run < tests/000.in > output.txt
      kks run -- arg_1 arg_2
    """

    directory = get_solution_directory()
    if directory is None:
        return

    binary = compile_solution(directory, mode)
    if binary is None:
        return

    output = f'Running binary'
    if len(run_args) > 0:
        output += ' with arguments ' + click.style(' '.join(run_args), fg='red', bold=True)
    click.secho(output, fg='green', err=True)

    subprocess.run([binary] + list(run_args), stdin=sys.stdin)



import subprocess

import click

from kks.util import get_solution_directory, get_clang_style_string


@click.command(short_help='Lint solution')
def lint():
    """
    Lint solution in current task directory using clang-format.

    clang-format has to be present in PATH.

    If file ~/.kks/.clang-format exists, it will be passed to clang-format.
    Otherwise, default config (hardcoded in util.py) is used.
    """

    directory = get_solution_directory()
    if directory is None:
        return

    c_files = list(directory.glob('*.c'))

    if len(c_files) == 0:
        click.secho('No *.c files found', fg='yellow', err=True)
        return

    file_names = [file.as_posix() for file in c_files]

    files_string = click.style(' '.join(file_names), fg='blue', bold=True)
    click.secho('Formatting files ' + files_string)

    style_string = '--style=' + get_clang_style_string()
    process = subprocess.run(['clang-format', '-i', style_string] + file_names)

    if process.returncode != 0:
        click.secho(f'Clang-format exited with exit-code {process.returncode}', fg='red', err=True)
        return

    click.secho(f'Successfully formatted!', fg='green', err=True)
